package ru.landgrafhomyak.serdha.api.v0.ddl

import kotlin.jvm.JvmName
import kotlin.jvm.JvmStatic

/**
 * Descriptor of a foreign key. Used for schema manipulations.
 *
 * @param ContainerTableUserExtension Type of [owner table's][ForeignKey.fromTable] user expression for static reporting errors when this descriptor passed to wrong table.
 * @param TargetTableUserExtension Type of [target table's][ForeignKey.toTable] user expression for static reporting errors when this descriptor passed to wrong table.
 */
public interface ForeignKey<ContainerTableUserExtension : Any, TargetTableUserExtension : Any> {
	/**
	 * Table that contains this foreign key. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("fromTable")
	public val fromTable: Table<ContainerTableUserExtension, *>

	/**
	 * Columns in [ForeignKey.fromTable] which are used to reference row in [ForeignKey.toTable]. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("fromColumns")
	public val fromColumns: List<Column<*, *, ContainerTableUserExtension>>

	/**
	 * Table referenced by this foreign key. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("toTable")
	public val toTable: Table<TargetTableUserExtension, *>

	/**
	 * Columns in [ForeignKey.toTable] that are identifying row(s). For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("toColumns")
	public val toColumns: List<Column<*, *, TargetTableUserExtension>>

	/**
	 * Enum of actions to do when value in any of [ForeignKey.toColumns] in referenced row is changed.
	 */
	public enum class OnUpdateAction {
		/**
		 * Don't perform any actions in [ForeignKey.fromColumns]. May break reference.
		 */
		NO_ACTION,

		/**
		 * Forbids changing values in [ForeignKey.toColumns] if there is at least one reference from [ForeignKey.fromTable].
		 */
		RESTRICT,

		/**
		 * Sets `null` value to all columns in [ForeignKey.fromColumns] if any cell in [ForeignKey.toColumns] in referenced row was changed.
		 *
		 * All of [ForeignKey.fromColumns] must be nullable.
		 */
		SET_NULL,

		/**
		 * Sets default value to all columns in [ForeignKey.fromColumns] if any cell in [ForeignKey.toColumns] in referenced row was changed.
		 *
		 * All of [ForeignKey.fromColumns] must have [DefaultConstraint].
		 */
		SET_DEFAULT,

		/**
		 * Reflects all changes from [ForeignKey.toColumns] to [ForeignKey.fromColumns].
		 */
		CASCADE
	}

	/**
	 * Enum of actions to do when referenced row in [ForeignKey.toTable] deleted.
	 */
	public enum class OnDeleteAction {
		/**
		 * Don't perform any actions in [ForeignKey.fromColumns]. Breaks reference.
		 */
		NO_ACTION,

		/**
		 * Forbids deleting row in [ForeignKey.toTable] if there is at least one reference from [ForeignKey.fromTable].
		 */
		RESTRICT,

		/**
		 * Sets `null` value to all columns in [ForeignKey.fromColumns] if referenced row was deleted.
		 *
		 * All of [ForeignKey.fromColumns] must be nullable.
		 */
		SET_NULL,

		/**
		 * Sets default value to all columns in [ForeignKey.fromColumns] if referenced row was deleted.
		 *
		 * All of [ForeignKey.fromColumns] must have [DefaultConstraint].
		 */
		SET_DEFAULT,

		/**
		 * Deletes all rows in [ForeignKey.fromTable] that are referencing deleted row in [ForeignKey.toTable].
		 */
		CASCADE
	}

	public companion object {
		@JvmStatic
		public val DEFAULT_ON_UPDATE_ACTION: OnUpdateAction
			inline get() = OnUpdateAction.CASCADE

		@JvmStatic
		public val DEFAULT_ON_DELETE_ACTION: OnDeleteAction
			inline get() = OnDeleteAction.RESTRICT
	}
}